/*************************************************************************
  SIM7600E TCPIP Library
  based on SIM800 GPRS/HTTP Library
  Distributed under GPL v2.0
  Written by Stanley Huang <stanleyhuangyc@gmail.com>
  For more information, please visit http://arduinodev.com

  Modified for use with SIM7600E

  V 1.0.0  TCP and TLS functions in one Lib
  V 1.1    22.03.2025 sendCommand buffers
  V 1.2    26.03.2025 modify init / start functions


*************************************************************************/

#include "SIM7600_beelogger.h"

extern uint8_t GSM_RX;

#include <AltSoftSerial.h>
AltSoftSerial SIM_SERIAL;       // RX, TX vom AT-Mega

const char SIM7600_UP[] PROGMEM    = "PB DONE"; // SIM7600 up and ready
const char SIM7600_CGD[] PROGMEM    = "AT+CGDCONT=1,\"IP\",\"";  // open IP-Socket
const char SIM7600_CGDREQ[] PROGMEM = "AT+CGDCONT?";
//TCP
const char SIM7600_NETOPN[] PROGMEM  = "AT+NETOPEN";
const char SIM7600_NETCLS[] PROGMEM  = "AT+NETCLOSE";
const char SIM7600_CIPMOD[] PROGMEM  = "AT+CIPMODE=0";
const char SIM7600_CIPOPN[] PROGMEM  = "AT+CIPOPEN=1,\"TCP\",\"";
const char SIM7600_CIPCLS[] PROGMEM  = "AT+CIPCLOSE=1";

//SSL
const char SIM7600_ADR[] PROGMEM     = "AT+CGPADDR";        // check local IP-Adress
const char SIM7600_MODE[] PROGMEM    = "AT+CCHMODE=1";        // select Mode Transparent
const char SIM7600_SET[] PROGMEM     = "AT+CCHSET=0,0";       // set Mode
const char SIM7600_START[] PROGMEM   = "AT+CCHSTART";       // Start SSL
const char SIM7600_CFG[] PROGMEM     = "AT+CCHSSLCFG=0,0";      // Set SSL Config to session
const char SIM7600_OPEN[] PROGMEM    = "AT+CCHOPEN=0,\"";     // Open Session to server
const char SIM7600_CON[] PROGMEM     = "CONNECT 96";        // check Connect 9600
const char SIM7600_CLOSE[] PROGMEM   = "AT+CCHCLOSE=0";       // Close Session
const char SIM7600_STOP[] PROGMEM    = "AT+CCHSTOP";        // Stop SSL

//location
const char SIM7600_NETSTA[] PROGMEM  = "AT+CNETSTART";
const char SIM7600_NETSTO[] PROGMEM  = "AT+CNETSTOP";

/**************************************************************************/
/*!
    @brief   setup Serial and check SIM7600E booting up
*/
/**************************************************************************/
bool CGPRS_SIM7600::init()
{
#define try_count 10
  char buf[16];
  uint8_t i = 0;

  SIM_SERIAL.begin(9600);

#if DEB_data
  Serial.println(F("SIM7600 26-03-2025"));
  Serial.flush();
#endif

  pinMode(GSM_RX, INPUT); // Turn pullup off  pinMode(INPUT_CAPTURE_PIN, INPUT_PULLUP);
  strcpy_P(buf, SIM7600_UP);
  do {  // Important, wait for SIM7600coming up
    if (sendCommand(0, 5000, buf)) {
      break;
    }
    i++;
  } while (i < try_count);
  if (i ==  try_count) return (false);
  
  i = 0;
  delay(1000);
  strcpy(buf, "ATE0");
  do {
    if (sendCommand(buf)) {
      break;
    }
    i++;
  }
  while (i < try_count);
  if (i ==  try_count) return (false);

  strcpy(buf, "AT+CFUN=1");
  if (sendCommand(buf, 10000)) { // SIM Card on
    delay(2000);
    return (true);
  }
  return false;
}

/**************************************************************************/
/*!
    @brief    turn off services and end Serial
*/
/**************************************************************************/
void CGPRS_SIM7600::shutdown()
{
  char buf[16] = "AT+CFUN=0";
  sendCommand(buf, 10000);
  strcpy(buf, "AT+CPOF");
  sendCommand(buf, 10000);
  SIM_SERIAL.end();  // stop listening
}


/**************************************************************************/
/*!
    @brief  SSL functions, transparent Mode
*/
/**************************************************************************/

/**************************************************************************/
/*!
    @brief   check network and connect to APN
*/
/**************************************************************************/
uint8_t CGPRS_SIM7600::start_TLS(const char* apn, const char* usr, const char* pw)
{
  bool success = false;
  uint8_t n = 0;
  char buf[64];
  strcpy(buf, "AT+CREG?");
  do {
    if (sendCommand(buf, 2000)) {  // are we registerd to a gsm-network
      char *p = strstr(buffer, "0,");
      if (p) {
        char mode = *(p + 2);
        if (mode == '1' || mode == '5') {  // net or roaming
          success = true;
          break;
        }
      }
    }
    delay(2000);
    n++;
  } while (n < 30);
  if (!success) return 1;  // not registered to network

  success = false;
  delay(1000);
  n = 0;
  strcpy(buf, "AT+CGREG?");
  do {
    if (sendCommand(buf, 2000)) {  // are we registerd to a gsm-network
      char *p = strstr(buffer, "0,");
      if (p) {
        char mode = *(p + 2);
        if (mode == '1' || mode == '5') {  // net or roaming
          success = true;
          break;
        }
      }
    }
    delay(2000);
    n++;
  } while (n < 5);
  if (!success) return 2;  // not registered to gprs network

  delay(1000);

  strcpy_P(buf, SIM7600_CGDREQ);
  sendCommand(buf, 5000);

  //activate context with APN
  strcpy_P(buf, SIM7600_CGD);
  strcat(buf, apn);
  strcat(buf, "\"");
  if (sendCommand(buf, 10000)) {
    delay(1000);  // wait on SIM7600 to get ready
#if 0
    if (strlen(usr) > 0) {
      strcpy(buf, "AT+CGAUTH=1,3,\"");
      strcat(buf, usr);
      strcat(buf, "\",\"");
      strcat(buf, pw);
      strcat(buf, "\"");
      sendCommand(buf, 5000);
      delay (500);
    }
#endif
    strcpy_P(buf, SIM7600_ADR);
    n = 0;
    do {
      if (sendCommand(buf, 5000, ".")) {  // +CGPADDR: 1,0.0.0.0, are we done internally?
        char *s = strstr(buffer, "1,");
        if (s) {
          char ip_ = *(s + 2);
          if (  ip_ != '0') { // ip adress starts with 1...9
            return (0);
          }
        }
      }
      delay(1000);
      n++;
    } while (n < 5);
    return (4); // no IP Adress
  }
  return (3); // no APN
}

/**************************************************************************/
/*!
    @brief   SSL functions, transparent Mode

*/
/**************************************************************************/
uint8_t CGPRS_SIM7600::ssl_init()
{
  char buf[16];
  strcpy_P(buf, SIM7600_MODE);
  if (sendCommand(buf, 5000)) { // 1 = transparent Mode
    delay(500);  // give SIM7600 some time to proceed
    strcpy_P(buf, SIM7600_SET);
    if (sendCommand(buf, 5000)) {
      delay(500);  // give SIM7600 some time to proceed
      return (1);
    }
  }
  return (0);
}
//start SSL, activate PDP context
uint8_t CGPRS_SIM7600::ssl_start()
{
  char buf[20];
  strcpy_P(buf, SIM7600_START);
  if (sendCommand(buf, 5000)) {
    delay(500);  // give SIM7600 some time to proceed
    strcpy_P(buf, SIM7600_CFG);
    if (sendCommand(buf, 5000)) {
      delay(500);  // give SIM7600 some time to proceed
      return (1);
    }
  }
  return (0);
}
//stop SSL
uint8_t CGPRS_SIM7600::ssl_stop()
{
  char buf[16];
  strcpy_P(buf, SIM7600_STOP);
  if (sendCommand(buf, 5000)) {
    delay(500);  // give SIM7600 some time to proceed
    return (1);
  }
  return (0);
}
//connect SSL
uint8_t CGPRS_SIM7600::ssl_open(const char* url)
{
  char buf[80];  // AT+CCHOPEN=0," server_name ",443,2
  char cmp[16];
  strcpy_P(cmp, SIM7600_CON); // CONNECT
  strcpy_P(buf, SIM7600_OPEN);
  strcat(buf, url);
  strcat(buf, "\",443,2");

  if (sendCommand(buf, 10000, cmp)) { // check CONNECT
    // Check CONNECT "FAIL"
    //strcpy(cmp, "FAIL");
    //if(sendCommand(0,1000,cmp) == 0){
    return (1); // FAIL not found -> success
    //}
  }
  return (0);
}
//close connection
uint8_t CGPRS_SIM7600::ssl_close()
{
  char buf[16];
  strcpy_P(buf, SIM7600_CLOSE);
  if (sendCommand(buf, 5000)) {
    return (1);
  }
  return (0);
}

//write data to module GET / ...
void CGPRS_SIM7600::ssl_data(const char* data)
{
#if DEB_data
  Serial.println(data);
  Serial.flush();
#endif
  SIM_SERIAL.print(data);
  SIM_SERIAL.flush();
}
// end TLS functions
//#####################################


//#####################################
// start TCP/IP functions
//#####################################
uint8_t CGPRS_SIM7600::start(const char* apn, const char* usr, const char* pw)
{
  bool success = false;
  uint8_t n = 0;
  char buf[64];
  strcpy(buf, "AT+CREG?");
  do {
    if (sendCommand(buf, 2000)) {  // are we registerd to a gsm-network
      char *p = strstr(buffer, "0,");   // expected 0,1 or 0,5
      if (p) {
        char mode = *(p + 2);
        if ((mode == '1' || mode == '5')) {  // net or roaming
          success = true;
          break;
        }
      }
    }
    delay(2000);
    n++;
  } while (n < 30);
  if (!success) return 1;

  delay(1000);
  n = 0;
  strcpy(buf, "AT+CGREG?");
  do {
    if (sendCommand(buf, 2000)) {  // are we registerd to a gsm-network
      char *p = strstr(buffer, "0,");    // expected 0,1 or 0,5
      if (p) {
        char mode = *(p + 2);
        if ((mode == '1' || mode == '5')) {// net or roaming
          success = true;
          break;
        }
      }
    }
    delay(2000);
    n++;
  } while (n < 5);
  if (!success) return 2;

  delay(1000);

  strcpy_P(buf, SIM7600_CIPMOD);
  sendCommand(buf, 5000);

  strcpy_P(buf, SIM7600_CGDREQ);
  sendCommand(buf, 5000);

#if 0
  if (strlen(usr) > 0) {
    strcpy(buf, "AT+CGAUTH=1,3,\"");
    strcat(buf, usr);
    strcat(buf, "\",\"");
    strcat(buf, pw);
    strcat(buf, "\"");
    sendCommand(buf, 5000);
    delay (500);
  }
#endif
  //activate PDP Context with APN
  strcpy_P(buf, SIM7600_CGD);
  strcat(buf, apn);
  strcat(buf, "\"");
  if (sendCommand(buf, 10000)) {
    n = 0;
    sendCommand("AT+CGACT=0,1");
    strcpy_P(buf, SIM7600_NETOPN);
    if (sendCommand(buf, 10000, ": 0")) { // activate Context +NETOPEN: 0
      strcpy(buf, "AT+IPADDR");
      do {
        if (sendCommand(buf, 10000, "0.")) { // are we done internally
          return (0);
        }
        delay(1000);
        n++;
      } while (n < 3);
      return (5);
    }
    return (4);
  }
  return (3);
}


//#####################################
//Deactivate Context
//makes context come back to original state
void CGPRS_SIM7600::stop()
{
  char buf[16];
  strcpy_P(buf, SIM7600_NETCLS);
  sendCommand(buf, 5000);
}

//#####################################
// CIP start connection to host
bool CGPRS_SIM7600::Connect(const char* host_name)
{
  char buf[64];
  strcpy_P(buf, SIM7600_CIPOPN);
  strcat(buf, host_name);
  strcat(buf, "\",80");       // port 80
  // set up or check for a connection +CIPOPEN: 1,0
  if (sendCommand(buf, 20000, "OPEN: 1,0")) {
    return true;
  }
  disConnect();
  return false;
}

//#####################################
//close TCP Connection
void CGPRS_SIM7600::disConnect()
{
  char buf[16];
  strcpy_P(buf, SIM7600_CIPCLS);
  sendCommand(buf, 5000, "1,0");
}


//#####################################
// prepare Sending data using CIPSEND
// and wait for being ready by returning ">"
bool CGPRS_SIM7600::prep_send()
{
  char buf[16] = "AT+CIPSEND=1,";
  if (sendCommand(buf, 10000, ">")) { // receive ready?
    return (true);
  }
  return (false);
}
//#####################################
// start sending data by calling function: send(0x00)
void CGPRS_SIM7600::send(const char* data)
{
  if (data == 0x00) {
    SIM_SERIAL.write(0x1A);       // transmission of data start
  }
  else {
#if DEB_data
    Serial.print(data);
    Serial.flush();
#endif
    SIM_SERIAL.print(data);
  }
}
// end TCP/IP functions
//#####################################



/**************************************************************************/
/*!
    @brief   get signal quality on the given entwork
*/
/**************************************************************************/
int CGPRS_SIM7600::getSignalQuality()
{
  char buf[8] = "AT+CSQ";
  sendCommand(buf, 10000, ","); // suche nach Komma zw. rssi und ber
  char *p = strstr(buffer, "CSQ: ");
  if (p) {
    int n = atoi(p + 5);
    if (n == 99 || n == -1) return -1;
    return n * 2 - 114;
  }
  return -2;
}
/**************************************************************************/
/*!
    @brief   get location data of the base station connected to
*/
/**************************************************************************/
// Location command:
//AT+CLBS=4
//OK
//+CLBS: 0,31.228525,121.380295,500,2025/06/07,10:49:08
//
bool CGPRS_SIM7600::getLocation(GSM_LOCATION * loc)
{
  bool x_val = false;
  char buf[16];
  strcpy_P(buf, SIM7600_NETSTA);
  if (sendCommand(buf, 15000)) {
    delay(1000);  // wait for SIM7600 to get ready
    strcpy(buf, "AT+CLBS=4");
    if (sendCommand(buf, 15000, "+CLBS")) {
      uint32_t t = millis();
      uint8_t n = 0;
      do { // read til EOL
        if (SIM_SERIAL.available()) {
          char c = SIM_SERIAL.read();
          buffer[n++] = c;
          if (c == '\r') {
            break; // EOL
          }
        }
      } while ((n < 50) && (millis() - t < 3000));
      buffer[n] = 0;
      char *p = buffer;
      do {
        if (!(p = strchr(p, ','))) break;
        loc->lon = atof(++p);
        if (!(p = strchr(p, ','))) break;
        loc->lat = atof(++p);
        /*
          if (!(p = strchr(p, ','))) break;
          loc->year = atoi(++p) - 2000;
          if (!(p = strchr(p, '/'))) break;
          loc->month = atoi(++p);
          if (!(p = strchr(p, '/'))) break;
          loc->day = atoi(++p);
          if (!(p = strchr(p, ','))) break;
          loc->hour = atoi(++p);
          if (!(p = strchr(p, ':'))) break;
          loc->minute = atoi(++p);
          if (!(p = strchr(p, ':'))) break;
          loc->second = atoi(++p);
        */
        x_val = true;
        break;
      } while (0);
    }
#if DEB_data
    Serial.print(buffer);
    Serial.flush();
#endif
    strcpy_P(buf, SIM7600_NETSTO);
    sendCommand(buf, 5000);
  }
  return x_val;
}

/**************************************************************************/
/*!
    @brief    Utility functions
*/
/**************************************************************************/
uint8_t CGPRS_SIM7600::sendCommand(const char* cmd, unsigned int timeout, const char* expected)
{
  if (cmd != 0) {
#if DEB_cmd
    Serial.print(F("\n**"));
    Serial.println(cmd);
    Serial.flush();
#endif
    purgeSerial();
    SIM_SERIAL.println(cmd);
  }
  uint32_t t = millis();
  uint8_t n = 0;
  char c;
  do {
    if (SIM_SERIAL.available()) {
      c = SIM_SERIAL.read();
      if (n >= (sizeof(buffer) - 1)) { // buffer full, discard first half
        n = (sizeof(buffer) / 2) - 1;
        memcpy(buffer, buffer + (sizeof(buffer) / 2), n);
      }
      buffer[n++] = c;
      buffer[n] = 0;
#if DEB_more
      Serial.print(c);
      //Serial.flush();
#endif
      if (strstr(buffer, expected ? expected : "OK\r")) {
        return n;
      }

      //if( c == '\r') n = 0; // do not use here, causes major trouble
    }
  } while (millis() - t < timeout);
#if DEB_more
  Serial.print(F("--t/o"));
  Serial.flush();
#endif
  return 0;
}

//#####################################
void CGPRS_SIM7600::purgeSerial()
{
  while (SIM_SERIAL.available()) {
    SIM_SERIAL.read();
    delay(5);
  }
}
